package kaka.cakelight;

import kaka.cakelight.mode.Mode;
import org.jnativehook.GlobalScreen;
import org.jnativehook.NativeHookException;
import org.jnativehook.keyboard.NativeKeyAdapter;
import org.jnativehook.keyboard.NativeKeyEvent;
import org.jnativehook.mouse.NativeMouseEvent;
import org.jnativehook.mouse.NativeMouseMotionAdapter;

public class CakeLight {
    private Configuration config;
    private Mode mode;
    private LedController ledController;

    public CakeLight(Configuration config, LedController ledController) {
        this.config = config;
        this.ledController = ledController;
        Color.calculateGammaCorrection(config.gamma);
    }

    public void setMode(Mode mode) {
        cleanup();
        this.mode = mode;
        mode.setFrameListener(ledController::onFrame);
        mode.enter(config);
    }

    public void cleanup() {
        if (this.mode != null) {
	    this.mode.setFrameListener(ledFrame -> {}); // To avoid any frame being sent to the controller while the thread is exiting
            this.mode.exit();
        }
    }

    public void startLoop() {
        Console console = Console.start(this, config);
        PipeController.start(console);
//        initNativeHook();
        // TODO
//        FrameGrabber grabber = FrameGrabber.from(config);
//        grabber.prepare();
//        VideoFrame frame = grabber.grabFrame();
//        double time = 0;
//        for (int i = 0; i < 100; i++) {
//            time += timeIt("frame", () -> grabber.grabFrame());
//        }
//        System.out.println("time = " + time);
//        grabber.close();
//	byte[] data = frame.getData();
//	saveFile(data, "/home/kaka/test.img");
    }

    private void initNativeHook() {
        try {
            GlobalScreen.registerNativeHook();
            GlobalScreen.addNativeKeyListener(new NativeKeyAdapter() {
                @Override
                public void nativeKeyPressed(NativeKeyEvent e) {
                    System.out.println("key code = " + e.getKeyCode() + ", key text = '" + NativeKeyEvent.getKeyText(e.getKeyCode()) + "'");
                }
            });
            GlobalScreen.addNativeMouseMotionListener(new NativeMouseMotionAdapter() {
                @Override
                public void nativeMouseMoved(NativeMouseEvent e) {
                    System.out.println("mouse point = " + e.getPoint());
                }
            });
        } catch (NativeHookException e) {
            e.printStackTrace();
        }
    }

    public void turnOff() {
        cleanup();
        ledController.onFrame(LedFrame.from(config).fillColor(0, 0, 0));
    }
}
